# -*- coding: utf-8 -*-

##############################################################################
# TTYLight-Python: easily stylize terminal text
##############################################################################
#
# Author: Travis Cardwell <travis.cardwell@yuzutechnology.com>
# Copyright: Copyright (c) 2013, Yuzu Technology, Inc.
#
# This file is subject to the terms and conditions defined in the License
# file included with this source code.
#
##############################################################################

"""
ttylight.ttylog Helpers
=======================

Helpers provide common terminal logging functionality for use in the
:mod:`ttylight.ttylog` logging library.  The following features are provided:

* displaying the current date/time
* displaying elapsed time
* displaying estimated remaining time
* displaying a counter
* displaying a progress counter
* displaying labels
* displaying "spinner" style animations

The :class:`LogHelper` class can be used to create customized log prefixes and
suffixes that combine any of the above features.  The :class:`Labeller`,
:class:`Spinner`, :class:`Counter`, :class:`ProgressCounter`,
:class:`Timestamper`, and :class:`Timer` subclasses provide easy APIs for
basic functionality.

See the source code of the :class:`LogHelper` class as a concrete example of
how to write your own log prefixes and suffixes.

API
---
"""


import datetime
import math

from ttylight import ttylog


def render_timedelta(delta, fmt=None):
    """Render a timedelta with the given format

    The following formats are supported:

    * 'S.UUU'
    * 'M:SS.UUU'
    * 'M:SS'
    * 'H:MM:SS'
    * 'DdHH:MM:SS'
    * 'DdHH:MM'

    :param datetime.timedelta delta: timedelta to render
    :param str fmt: format (default/invalid: 'H:MM:SS')
    :returns: the rendered timedelta
    :rtype: str
    """
    tsecs = delta.total_seconds()
    if fmt == 'S.UUU':
        return '{:0.03f}'.format(tsecs)
    if fmt == 'M:SS.UUU':
        mins, _secs = divmod(math.floor(tsecs), 60)
        secs = tsecs - 60 * mins
        return '{:d}:{:06.03f}'.format(mins, secs)
    mins, secs = divmod(int(round(tsecs)), 60)
    if fmt == 'M:SS':
        return '{:d}:{:02d}'.format(mins, secs)
    hours, mins = divmod(mins, 60)
    if fmt == 'H:MM:SS':
        return '{:d}:{:02d}:{:02d}'.format(hours, mins, secs)
    days, hours = divmod(hours, 24)
    if fmt == 'DdHH:MM:SS':
        return '{:d}d{:02d}:{:02d}:{:02d}'.format(days, hours, mins, secs)
    if fmt == 'DdHH:MM':
        return '{:d}d{:02d}:{:02d}'.format(days, hours, mins)
    return '{:d}:{:02d}:{:02d}'.format(days * 24 + hours, mins, secs)


class LogHelper:
    """log helper

    This class implements many common logging utilities.  Instances are meant
    to be used as a prefix or suffix in functions of the
    :mod:`ttylight.ttylog` module.

    The following features are supported:

    * 'current' - the current date/time
    * 'elapse' - the elapsed time
    * 'remain' - the remaining time (estimate)
    * 'counter' - the current counter
    * 'end' - the counter end
    * 'label' - the current label
    * 'state' - the current spinner state

    The output is specified in a template string, which is formatted using
    `standard string formatting
    <http://docs.python.org/3.3/library/string.html#string-formatting>`_.
    For example, the following template right-aligns the label within 12
    characters and shows a progress counter::

        from ttylight.ttylog.helpers import LogHelper

        lh = LogHelper('{label: >12} [{counter}/{end}]', end=100, label="A")

    Note that each field in the template is a string.  The ``counter`` and
    ``end`` fields are rendered as strings before being formatted in the
    template.  The ``width`` parameter allows you to specify the width at
    which these integral fields are rendered.

    For an accurate ``remain`` time estimate, the counter must be maintained
    by using :meth:`disp_inc`, :meth:`inc_disp`, or :meth:`inc`, but it is not
    necessary to use ``counter`` in the template.

    The ``dt_fmt`` parameter is used to specify the format in which the
    ``current`` date/time is rendered, using
    `datetime.datetime.strftime
    <http://docs.python.org/3.3/library/datetime.html
    #strftime-and-strptime-behavior>`_.
    The default format is ISO 8601.

    The ``elapse_fmt`` and ``remain_fmt`` parameters are used to specify the
    formats of the ``elapse`` time and ``remain`` time, respectively.  The
    valid formats are defined by the :func:`render_timedelta` function.  The
    default format for both is 'H:MM:SS'.

    The ``tz`` parameter is used to specify a
    `time zone
    <http://docs.python.org/3.3/library/datetime.html#tzinfo-objects>`_.
    Local time is used by default.

    The ``label`` parameter specifies the initial label.  An empty string is
    used by default.

    The ``states`` parameter specifies the states that should be looped
    through when displaying.  If each state is one character long, then the
    parameter can be specified as a string.  If a state is more than one
    character long, then the parameter must be specified as a list of strings.
    For example, the following code results in a "spinner" animation::

        from ttylight.ttylog.helpers import LogHelper

        lh = LogHelper('{state}', states='|/-\\\\', sameline=True)

    The ``sameline`` parameter causes the helper to output an ANSI sequence
    that results in multiple lines of the output (printed consecutively) to
    be printed on the same line.  This feature should only be used with log
    prefixes.

    Using this class directly allows you to mix and match all of the above
    features.  The following subclasses are available as well, however, to
    provide an easy API for common functionality:

    * :class:`Labeller`
    * :class:`Spinner`
    * :class:`Counter`
    * :class:`ProgressCounter`
    * :class:`Timestamper`
    * :class:`Timer`

    See the source code of these subclasses as concrete examples of how to
    instantiate this class.
    """

    def __init__(self, template, start=1, end=1, width=None,
                 dt_fmt='%Y-%m-%dT%H:%M:%S%z',
                 elapse_fmt='H:MM:SS', remain_fmt='H:MM:SS', tz=None,
                 label='', states=None, sameline=False):
        """Initialize a LogHelper

        :param str template: display template
        :param int start: starting count, >= 0 (default: 1)
        :param int end: ending count, >= start (default: 1)
        :param int width: width of counter, >= 0 (default: <auto>)
        :param str dt_fmt: current time format (default: ISO 8601)
        :param str elapse_fmt: elapsed format (default: 'H:MM:SS')
        :param str remain_fmt: remain format (default: 'H:MM:SS')
        :param datetime.tzinfo tz: time zone (default: local)
        :param str label: initial label (default: '')
        :param [str] states: spinner states (default: [''])
        :param bool sameline: True to display on single line (default: False)
        """
        self._template = template
        self._count = max(start, 0)
        self._end = max(end, start)
        if width is None:
            self._width = math.floor(math.log10(self._end)) + 1
        else:
            self._width = max(width, 0)
        self._dt_fmt = dt_fmt
        self._elapse_fmt = elapse_fmt
        self._remain_fmt = remain_fmt
        self._tz = tz
        self._label = label
        if states is None:
            self._states = ['']
            self._states_len = 1
        else:
            self._states = states
            self._states_len = len(states)
        self._states_i = 0
        self._sameline = sameline and ttylog.State._ansi
        self._first = True
        self._render_current = template.find('current') >= 0
        self._render_elapse = template.find('elapse') >= 0
        self._render_remain = template.find('remain') >= 0
        self._render_counter = template.find('counter') >= 0
        self._render_end = template.find('end') >= 0
        self._render_label = template.find('label') >= 0
        self._render_state = template.find('state') >= 0
        self._dt_start = datetime.datetime.now(self._tz)

    def __repr__(self):
        """Render the LogHelper

        :returns: the current LogHelper state
        :rtype: str
        """
        parts = {}
        dt_cur = datetime.datetime.now(self._tz)
        if self._render_current:
            parts['current'] = dt_cur.strftime(self._dt_fmt)
        if self._render_elapse:
            parts['elapse'] = render_timedelta(
                dt_cur - self._dt_start, self._elapse_fmt)
        if self._render_remain:
            if self._count > 0:
                elapsed = dt_cur - self._dt_start
                remain = elapsed.total_seconds()
                remain = datetime.timedelta(
                    seconds=(remain * self._end / self._count - remain))
            else:
                remain = datetime.timedelta(seconds=0)
            parts['remain'] = render_timedelta(remain, self._remain_fmt)
        if self._render_counter:
            parts['counter'] = '{:0{w}d}'.format(self._count, w=self._width)
        if self._render_end:
            parts['end'] = '{:0{w}d}'.format(self._end, w=self._width)
        if self._render_label:
            parts['label'] = self._label
        if self._render_state:
            parts['state'] = self._states[self._states_i]
            self._states_i += 1
            if self._states_i >= self._states_len:
                self._states_i = 0
        s = ''
        if self._sameline:
            if self._first:
                self._first = False
            else:
                s = '\033[F\033[K'
        return s + self._template.format(**parts)

    def reset(self, count=1):
        """Reset the LogHelper state

        This method sets the ``count`` to the value specified by the argument
        of the same name, causes the ``state`` to start over at the beginning
        of the states list, forces the next output to be printed on a new
        line, and resets the starting timestamp that is used to calculate the
        ``remain`` time.

        :param int count: reset count (default: 1)
        """
        self._count = count
        self._states_i = 0
        self._first = True
        self._dt_start = datetime.datetime.now(self._tz)

    def inc(self, count=1):
        """Increment the LogHelper counter

        Use this function to increment the count by hand.

        :param int count: number to increment by (default: 1)
        """
        self._count += count

    def disp_inc(self):
        """Get the LogHelper state and then increment the counter

        Set this function as a prefix or suffix of a log function in order to
        have the count automatically increment after each time it is
        displayed.

        :returns: LogHelper state before incrementing
        :rtype: str
        """
        s = str(self)
        self._count += 1
        return s

    def inc_disp(self):
        """Get the LogHelper state after incrementing the counter

        Set this function as a prefix or suffix of a log function in order to
        have the count automatically increment before each time it is
        displayed.

        :returns: LogHelper state after incrementing
        :rtype: str
        """
        self._count += 1
        return str(self)

    def set_label(self, label):
        """Set the label

        :param str label: new label
        """
        self._label = label


class Labeller(LogHelper):
    """LogHelper that displays labels

    This is a subclass of :class:`LogHelper` that provides an easy API for
    labelling logs.

    The initial label is set using the ``label`` parameter of the constructor.
    The label can then be changed at any time using the
    :meth:`set_label <LogHelper.set_label>` method.

    The ``align`` parameter sets the alignment of the labels within a field
    width specified by the ``width`` parameter.  ``align`` should be one of
    the alignment class variable constants defined in this class.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the label field.  Empty strings are
    used by default.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Here is an example::

        from ttylight import ttylog
        from ttylight.ttylog.helpers import Labeller

        labeller = Labeller('low', align=Labeller.RIGHT, width=6, suffix=':')
        ttylog.State.customize(ttylog.State.CUSTOM, prefix=labeller)
        ttylog.custom('1')
        ttylog.custom('2')
        labeller.set_label('mid')
        ttylog.custom('3')
        ttylog.custom('4')
        labeller.set_label('high')
        ttylog.custom('5')
        ttylog.custom('6')

    :var int LEFT: left alignment constant
    :var int CENTER: center alignment constant
    :var int RIGHT: right alignment constant
    """

    LEFT = 0
    CENTER = 1
    RIGHT = 2

    def __init__(self, label, align=0, width=None, prefix='', suffix='',
                 sameline=False):
        """Initialize a Labeller

        :param str label: initial label
        :param str align: {LEFT|CENTER|RIGHT} (default: LEFT)
        :param int width: label field width (default: <auto>)
        :param str prefix: prefix string (default: '')
        :param str suffix: suffix string (default: '')
        :param bool sameline: True to display on single line (default: False)
        """
        template = prefix.replace('{', '{{').replace('}', '}}') + '{label'
        if width is None:
            template += '}'
        else:
            template += (': ' + {0: '<', 1: '^', 2: '>'}.get(align, '<') +
                         str(width) + '}')
        template += suffix.replace('{', '{{').replace('}', '}}')
        super().__init__(template, label=label, sameline=sameline)


class Spinner(LogHelper):
    """LogHelper that displays ANSI animations

    This is a subclass of :class:`LogHelper` that provides an easy API for
    adding silly animations to logs.  As ``sameline`` is necessary for the
    animation to work, this helper is generally used as a prefix.  It can
    be used as a suffix if logs are of fixed width, ``sameline=True`` for the
    log prefix, and ``sameline=False`` for the log suffix.

    The ``animation`` parameter sets the type of animation to use, specified
    by the animation class variable constants defined in this class.

    The ``width`` parameter sets the width of the animation, in characters.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the animation.  By default, the
    animation is enclosed in parenthesis.

    The ``sameline`` parameter is described in :class:`LogHelper`.  It is on
    by default, in order to make an animation.

    Here is an example::

        import time
        from ttylight import ttylog
        from ttylight.ttylog.helpers import Spinner

        ttylog.State.customize(ttylog.State.CUSTOM,
                               prefix=Spinner(Spinner.SPINNER_C))
        for i in range(1, 33):
            ttylog.custom(str(i))
            time.sleep(0.2)

    :var int SPINNER_C: spinner (clockwise) animation constant
    :var int SPINNER_CC: spinner (counter-clockwise) animation constant
    :var int PULSER: pulser animation constant
    :var int PACER: pacer animation constant
    :var int PINBALL: pinball animation constant
    """

    SPINNER_C = 0
    SPINNER_CC = 1
    PULSER = 2
    PACER = 3
    PINBALL = 4

    def __init__(self, animation=0, width=1, prefix='(', suffix=')',
                 sameline=True):
        """Initialize a Spinner

        :param int animation: {SPINNER_C|SPINNER_CC|PULSER|PACER|PINBALL}
        :param int width: animation width (default: 1)
        :param str prefix: prefix string (default: '(')
        :param str suffix: suffix string (default: ')')
        :param bool sameline: True to display on single line (default: False)
        """
        states = []
        if animation == Spinner.PINBALL:
            if width <= 1:
                states = '|\\'
            elif width == 2:
                states = ['||', '\\/']
            else:
                w3 = width - 3
                states.append('\\*' + ' ' * w3 + '|')
                for i in range(1, w3):
                    states.append('|' + ' ' * i + '*' + ' ' * (w3 - i) + '|')
                states.append('|' + ' ' * w3 + '*/')
                for i in range(1, w3):
                    states.append('|' + ' ' * (w3 - i) + '*' + ' ' * i + '|')
        elif animation == Spinner.PACER:
            if width <= 1:
                states = '|>|<'
            elif width == 2:
                states = ['| ', '> ', ' |', ' <']
            elif width == 3:
                states = ['|  ', ' > ', '  |', ' < ']
            else:
                w2 = width - 2
                states.append('|' + ' ' * w2 + ' ')
                for i in range(1, w2):
                    states.append(' ' * i + '>' + ' ' * (w2 - i) + ' ')
                states.append(' ' + ' ' * w2 + '|')
                for i in range(1, w2):
                    states.append(' ' + ' ' * (w2 - i) + '<' + ' ' * i)
        else:
            frames = {0: '|/-\\',
                      1: '|\\-/',
                      2: '.oOo'}.get(animation, '|/-\\')
            states = [c for c in frames]
            for i in range(1, width):
                ss = []
                for c in frames:
                    ss += [c + s for s in states]
                states = ss
        template = (prefix.replace('{', '{{').replace('}', '}}') + '{state}' +
                    suffix.replace('{', '{{').replace('}', '}}'))
        super().__init__(template, states=states, sameline=sameline)


class Counter(LogHelper):
    """LogHelper that displays a counter

    This is a subclass of :class:`LogHelper` that provides an easy API for
    adding a counter to logs.

    The ``start`` parameter can be used to set the starting count.  It is set
    to 1 by default.

    The ``width`` parameter can be used to set the width of the counter,
    padding it with zeros.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the counter.  By default, the counter
    is enclosed in brackets.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Common usage of this class involves passing the
    :meth:`disp_inc <LogHelper.disp_inc>` method as the log prefix or log
    suffix, which causes the counter to be incremented after each time it is
    displayed.  Alternatively, the class can be instantiated with ``start=0``,
    and the :meth:`inc_disp <LogHelper.inc_disp>` method can be passed as the
    log prefix or log suffix, which causes the counter to be incremented
    before each time it is displayed.  The counter can also be incremented
    manually using the :meth:`inc <LogHelper.inc>` method, in which case the
    instance itself is passed as the log prefix or log suffix.

    Here is an example::

        from ttylight import ttylog
        from ttylight.ttylog.helpers import Counter

        counter = Counter(width=2)
        ttylog.State.customize(ttylog.State.CUSTOM, prefix=counter)
        for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            ttylog.custom(c)
    """

    def __init__(self, start=1, width=None, prefix='[', suffix=']',
                 sameline=False):
        """Initialize a Counter

        :param int start: starting count (default: 1)
        :param int width: width of counter (default: <auto>)
        :param str prefix: prefix string (default: '[')
        :param str suffix: suffix string (default: ']')
        :param bool sameline: True to display on single line (default: False)
        """
        template = (prefix.replace('{', '{{').replace('}', '}}') +
                    '{counter}' +
                    suffix.replace('{', '{{').replace('}', '}}'))
        super().__init__(template, start=start, width=width,
                         sameline=sameline)


class ProgressCounter(LogHelper):
    """LogHelper that displays a progress counter

    This is a subclass of :class:`LogHelper` that provides an easy API for
    adding a progress counter to logs.

    The ``start`` parameter can be used to set the starting count.  It is set
    to 1 by default.

    The ``end`` parameter can be used to set the end count.  It is set to 100
    by default.

    The ``width`` parameter can be used to set the width of ``count`` and
    ``end``, padding them with zeros.  By default, ``width`` is automatically
    set to the number of digits in ``end``.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the progress counter.  By default, the
    progress counter is enclosed in brackets.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Common usage of this class involves passing the
    :meth:`disp_inc <LogHelper.disp_inc>` method as the log prefix or log
    suffix, which causes the counter to be incremented after each time it is
    displayed.  Alternatively, the class can be instantiated with ``start=0``,
    and the :meth:`inc_disp <LogHelper.inc_disp>` method can be passed as the
    log prefix or log suffix, which causes the counter to be incremented
    before each time it is displayed.  The counter can also be incremented
    manually using the :meth:`inc <LogHelper.inc>` method, in which case the
    instance itself is passed as the log prefix or log suffix.

    Here is an example::

        from ttylight import ttylog
        from ttylight.ttylog.helpers import ProgressCounter

        counter = Counter(end=26)
        ttylog.State.customize(ttylog.State.CUSTOM, prefix=counter)
        for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ':
            ttylog.custom(c)
    """

    def __init__(self, start=1, end=100, width=None, prefix='[', suffix=']',
                 sameline=False):
        """Initialize a ProgressCounter

        :param int start: starting count (default: 1)
        :param int end: ending count (default: 100)
        :param int width: width of counter (default: <auto>)
        :param str prefix: prefix string (default: '[')
        :param str suffix: suffix string (default: ']')
        :param bool sameline: True to display on single line (default: False)
        """
        template = (prefix.replace('{', '{{').replace('}', '}}') +
                    '{counter}' + '/' + '{end}' +
                    suffix.replace('{', '{{').replace('}', '}}'))
        super().__init__(template, start=start, end=end, width=width,
                         sameline=sameline)


class Timestamper(LogHelper):
    """LogHelper that displays timestamps

    This is a subclass of :class:`LogHelper` that provides an easy API for
    adding timestamps to logs.

    The ``dt_fmt`` parameter is used to specify the format in which the
    date/time is rendered, using
    `datetime.datetime.strftime
    <http://docs.python.org/3.3/library/datetime.html
    #strftime-and-strptime-behavior>`_.
    The default format is ISO 8601.

    The ``tz`` parameter is used to specify a
    `time zone
    <http://docs.python.org/3.3/library/datetime.html#tzinfo-objects>`_.
    Local time is used by default.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the progress counter.  Empty strings
    are used by default.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    Here is an example::

        import time
        from ttylight import ttylog
        from ttylight.ttylog.helpers import Timestamper

        ttylog.State.customize(ttylog.State.CUSTOM, prefix=Timestamper())
        for i in range(1, 33):
            ttylog.custom(str(i))
            time.sleep(0.2)
    """

    def __init__(self, dt_fmt='%Y-%m-%dT%H:%M:%S%z', tz=None, prefix='',
                 suffix='', sameline=False):
        """Initialize a Timestamper

        :param str dt_fmt: current time format (default: ISO 8601)
        :param datetime.tzinfo tz: time zone (default: local)
        :param str prefix: prefix string (default: '[')
        :param str suffix: suffix string (default: ']')
        :param bool sameline: True to display on single line (default: False)
        """
        template = (prefix.replace('{', '{{').replace('}', '}}') +
                    '{current}' +
                    suffix.replace('{', '{{').replace('}', '}}'))
        super().__init__(template, dt_fmt=dt_fmt, tz=tz, sameline=sameline)


class Timer(LogHelper):
    """LogHelper that displays a progress timer

    This is a subclass of :class:`LogHelper` that provides an easy API for
    adding progress timers to logs.

    The ``delta_fmt`` parameter is used to specify the format of the
    ``elapse`` and ``remain`` times.  The valid formats are defined by the
    :func:`render_timedelta` function.  The default format for both is
    'H:MM:SS'.

    The ``start`` parameter can be used to set the starting count.  It is set
    to 1 by default.

    The ``end`` parameter can be used to set the end count.  It is set to 100
    by default.

    The ``prefix`` and ``suffix`` parameters are used to set strings that
    should be appended and prepended to the progress counter.  Empty strings
    are used by default.

    The ``sameline`` parameter is described in :class:`LogHelper`.

    For an accurate ``remain`` time estimate, the counter must be maintained.
    Common usage of this class involves passing the
    :meth:`disp_inc <LogHelper.disp_inc>` method as the log prefix or log
    suffix, which causes the counter to be incremented after each time it is
    displayed.  Alternatively, the class can be instantiated with ``start=0``,
    and the :meth:`inc_disp <LogHelper.inc_disp>` method can be passed as the
    log prefix or log suffix, which causes the counter to be incremented
    before each time it is displayed.  The counter can also be incremented
    manually using the :meth:`inc <LogHelper.inc>` method, in which case the
    instance itself is passed as the log prefix or log suffix.

    Here is an example::

        import time
        from ttylight import ttylog
        from ttylight.ttylog.helpers import Timer

        ttylog.State.customize(ttylog.State.CUSTOM, prefix=Timer(end=32))
        for i in range(1, 33):
            ttylog.custom(str(i))
            time.sleep(0.2)
    """

    def __init__(self, delta_fmt='H:MM:SS', start=1, end=100, prefix='',
                 suffix='', sameline=False):
        """
        :param str delta_fmt: time format (default: 'H:MM:SS')
        :param int start: starting count (default: 1)
        :param int end: ending count (default: 100)
        :param str prefix: prefix string (default: '[')
        :param str suffix: suffix string (default: ']')
        :param bool sameline: True to display on single line (default: False)
        """
        template = (prefix.replace('{', '{{').replace('}', '}}') +
                    '{remain}' + '->' + '{elapse}' +
                    suffix.replace('{', '{{').replace('}', '}}'))
        super().__init__(template, elapse_fmt=delta_fmt, remain_fmt=delta_fmt,
                         start=start, end=end, sameline=sameline)
